# sage_setup: distribution = sagemath-singular
"""
Arithmetic Engine for Polynomials as Tuples
"""
# ****************************************************************************
#  Copyright (C) 2021 Guillermo Aboumrad <gh_willieab>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

###########
#   API   #
###########

cpdef inline tuple poly_to_tup(MPolynomial_libsingular poly):
    r"""
    Convert a polynomial object into the internal representation as tuple of
    ``(ETuple exp, NumberFieldElement coeff)`` pairs.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: R.<x, y> = PolynomialRing(QQ)
        sage: poly_to_tup(x**2 + 1)
        (((2, 0), 1), ((0, 0), 1))
        sage: poly_to_tup(x**2*y**4 - 4/5*x*y**2 + 1/3 * y)
        (((2, 4), 1), ((1, 2), -4/5), ((0, 1), 1/3))
    """
    return tuple(poly.monomial_coefficients().items())

cpdef inline MPolynomial_libsingular _tup_to_poly(tuple eq_tup, MPolynomialRing_libsingular parent):
    r"""
    Return a polynomial object from its tuple of pairs representation.

    Inverse of :meth:`poly_to_tup`:

    - ``poly_to_tup(tup_to_poly(eq_tup, ring)) == eq_tup`` and
    - ``tup_to_poly(poly_to_tup(eq), eq.parent()) == eq``.

    .. NOTE::

        Assumes ``all(parent.ngens() == len(exp_tup) for exp_tup, c in eq_tup)``.
        This method is meant for internal use.

    .. WARNING::

      Unsafe for client use, since it avoids implicit casting and
      it may lead to segmentation faults.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.poly_tup_engine import _tup_to_poly
        sage: K = CyclotomicField(20)
        sage: R.<x, y> = PolynomialRing(K)
        sage: poly_tup = (((2, 0), K.one()), ((0, 0), K.one()))
        sage: _tup_to_poly(poly_tup, parent=R)
        x^2 + 1
        sage: poly = x**2*y**4 - 4/5*x*y**2 + 1/3 * y
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: _tup_to_poly(poly_to_tup(poly), parent=R) == poly
        True

    TESTS::

        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup, _tup_to_poly
        sage: R.<x, y, z> = PolynomialRing(CyclotomicField(20))
        sage: r = R.random_element()
        sage: _tup_to_poly(poly_to_tup(r), parent=R) == r
        True
        sage: R = PolynomialRing(QQ, 'fx', 100)
        sage: r = R.random_element()
        sage: _tup_to_poly(poly_to_tup(r), parent=R) == r
        True
    """
    return parent._element_constructor_(dict(eq_tup), check=False)

cdef inline tuple _flatten_coeffs(tuple eq_tup):
    r"""
    Flatten cyclotomic coefficients to a representation as a tuple of rational
    coefficients.

    This is used to avoid pickling cyclotomic coefficient objects, which fails
    with new PARI settings introduced in :issue:`30537`.
    """
    cdef list flat = []
    cdef NumberFieldElement_absolute cyc_coeff
    for exp, cyc_coeff in eq_tup:
        flat.append((exp, tuple(cyc_coeff._coefficients())))
    return tuple(flat)

cpdef tuple _unflatten_coeffs(field, tuple eq_tup):
    r"""
    Restore cyclotomic coefficient object from its tuple of rational
    coefficients representation.

    Used to circumvent pickling issue introduced by PARI settings
    in :issue:`30537`.

    EXAMPLES::

        sage: # needs sage.combinat sage.groups
        sage: from sage.algebras.fusion_rings.poly_tup_engine import _unflatten_coeffs
        sage: fm = FusionRing("A2", 2).get_fmatrix()
        sage: p = fm._poly_ring.random_element()
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: flat_poly_tup = list()
        sage: for exp, cyc_coeff in poly_to_tup(p):
        ....:     flat_poly_tup.append((exp, tuple(cyc_coeff._coefficients())))
        sage: flat_poly_tup = tuple(flat_poly_tup)
        sage: _unflatten_coeffs(fm.field(), flat_poly_tup) == poly_to_tup(p)
        True
    """
    cdef list unflat = []
    for exp, coeff_tup in eq_tup:
        unflat.append((exp, field(list(coeff_tup))))
    return tuple(unflat)

#################################
#   Useful private predicates   #
#################################

cdef inline int has_appropriate_linear_term(tuple eq_tup) noexcept:
    r"""
    Determine whether the given tuple of pairs (of length 2) contains
    an *appropriate* linear term.

    In this context, a linear term is said to be *appropriate* if
    it is in the largest variable in the given polynomial (w.r.t.
    the degrevlex ordering), the monomial in which the linear term
    appears is univariate, and the linear term is not a common factor in
    the polynomial.

    OUTPUT:

    If the given polynomial contains an appropriate linear term, this method
    returns the index of the monomial in which the term appears.

    Otherwise, the method returns `-1`.
    """
    max_var = degrees(eq_tup).nonzero_positions()[0]
    cdef ETuple m
    cdef int i
    for i in range(2):
        m = eq_tup[i][0]
        if m._nonzero == 1 and m._data[1] == 1 and m._data[0] == max_var and eq_tup[(i+1) % 2][0][max_var] == 0:
            return i
    return -1

######################
#   "Change rings"   #
######################

cpdef inline tup_to_univ_poly(tuple eq_tup, univ_poly_ring):
    r"""
    Given a tuple of pairs representing a univariate polynomial and a univariate
    polynomial ring, return a univariate polynomial object.

    Each pair in the tuple is assumed to be of the form ``(ETuple, coeff)``,
    where ``coeff`` is an element of ``univ_poly_ring.base_ring()``.

    EXAMPLES::

      sage: from sage.algebras.fusion_rings.poly_tup_engine import tup_to_univ_poly
      sage: from sage.rings.polynomial.polydict import ETuple
      sage: K = CyclotomicField(56)
      sage: poly_tup = ((ETuple([0, 3, 0]), K(2)), (ETuple([0, 1, 0]), K(-1)), (ETuple([0, 0, 0]), K(-2/3)))
      sage: R = K['b']
      sage: tup_to_univ_poly(poly_tup, R)
      2*b^3 - b - 2/3

    TESTS::

      sage: poly_tup = ((ETuple([0, 0, 0]), K(-1/5)), )
      sage: tup_to_univ_poly(poly_tup, R)
      -1/5
    """
    cdef ETuple exp
    cdef NumberFieldElement_absolute c
    return univ_poly_ring({exp._data[1] if exp._nonzero else 0: c for exp, c in eq_tup})

cpdef inline tuple resize(tuple eq_tup, dict idx_map, int nvars):
    r"""
    Return a tuple representing a polynomial in a ring with
    ``len(sorted_vars)`` generators.

    This method is used for creating polynomial objects with the
    "right number" of variables for computing Groebner bases of the
    partitioned equations graph and for adding constraints ensuring certain
    F-symbols are nonzero.

    EXAMPLES::

      sage: from sage.algebras.fusion_rings.poly_tup_engine import resize
      sage: from sage.rings.polynomial.polydict import ETuple
      sage: K = CyclotomicField(56)
      sage: poly_tup = ((ETuple([0, 3, 0, 2]), K(2)), (ETuple([0, 1, 0, 1]), K(-1)), (ETuple([0, 0, 0, 0]), K(-2/3)))
      sage: idx_map = {1: 0, 3: 1}
      sage: resize(poly_tup, idx_map, 2)
      (((3, 2), 2), ((1, 1), -1), ((0, 0), -2/3))

      sage: R = PolynomialRing(K, 'fx', 20)
      sage: R.inject_variables()
      Defining fx0, fx1, fx2, fx3, fx4, fx5, fx6, fx7, fx8, fx9, fx10, fx11, fx12, fx13, fx14, fx15, fx16, fx17, fx18, fx19
      sage: sparse_poly = R(fx0**2 * fx17 + fx3)
      sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup, _tup_to_poly
      sage: S.<x, y, z> = PolynomialRing(K)
      sage: _tup_to_poly(resize(poly_to_tup(sparse_poly), {0:0, 3:1, 17:2}, 3), parent=S)
      x^2*z + y
    """
    cdef ETuple exp, new_e
    cdef NumberFieldElement_absolute c
    cdef list resized = list()
    for exp, c in eq_tup:
        new_e = ETuple({idx_map[pos]: d for pos, d in exp.sparse_iter()}, nvars)
        resized.append((new_e, c))
    return tuple(resized)

###########################
#   Convenience methods   #
###########################

cdef inline ETuple degrees(tuple poly_tup):
    r"""
    Return the maximal degree of each variable in the polynomial.
    """
    # Deal with the empty tuple, representing the zero polynomial
    if not poly_tup:
        return ETuple()
    cdef ETuple max_degs
    cdef int i
    max_degs = <ETuple> (<tuple> poly_tup[0])[0]
    for i in range(1, len(poly_tup)):
        max_degs = max_degs.emax(<ETuple> (<tuple> poly_tup[i])[0])
    return max_degs

cpdef list get_variables_degrees(list eqns, int nvars):
    r"""
    Find maximum degrees for each variable in equations.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.poly_tup_engine import get_variables_degrees
        sage: R.<x, y, z> = PolynomialRing(QQ)
        sage: polys = [x**2 + 1, x*y*z**2 - 4*x*y, x*z**3 - 4/3*y + 1]
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: get_variables_degrees([poly_to_tup(p) for p in polys], 3)
        [2, 1, 3]
    """
    if not eqns:
        return [0]*nvars
    cdef ETuple max_deg
    cdef int i
    max_deg = degrees(eqns[0])
    for i in range(1, len(eqns)):
        max_deg = max_deg.emax(degrees(<tuple>(eqns[i])))
    cdef list dense = [0] * len(max_deg)
    for i in range(max_deg._nonzero):
        dense[max_deg._data[2*i]] = max_deg._data[2*i+1]
    return dense

cpdef list variables(tuple eq_tup):
    """
    Return indices of all variables appearing in ``eq_tup``.

    EXAMPLES::

      sage: from sage.algebras.fusion_rings.poly_tup_engine import variables
      sage: from sage.rings.polynomial.polydict import ETuple
      sage: poly_tup = ((ETuple([0, 3, 0]), 2), (ETuple([0, 1, 0]), -1), (ETuple([0, 0, 0]), -2/3))
      sage: variables(poly_tup)
      [1]
      sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
      sage: R.<x, y, z> = PolynomialRing(QQ)
      sage: variables(poly_to_tup(x*2*y + y**3 - 4/3*x))
      [0, 1]
      sage: variables(poly_to_tup(R(1/4)))
      []
    """
    return degrees(eq_tup).nonzero_positions()

cpdef constant_coeff(tuple eq_tup, field):
    r"""
    Return the constant coefficient of the polynomial represented by
    given tuple.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.poly_tup_engine import constant_coeff
        sage: from sage.rings.polynomial.polydict import ETuple
        sage: poly_tup = ((ETuple([0, 3, 0]), 2), (ETuple([0, 1, 0]), -1), (ETuple([0, 0, 0]), -2/3))
        sage: constant_coeff(poly_tup, QQ)
        -2/3
        sage: R.<x, y, z> = PolynomialRing(QQ)
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: constant_coeff(poly_to_tup(x**5 + x*y*z - 9), QQ)
        -9
    """
    cdef ETuple exp
    for exp, coeff in eq_tup:
        if exp.is_constant():
            return coeff
    return field.zero()

cpdef tuple apply_coeff_map(tuple eq_tup, coeff_map):
    """
    Apply ``coeff_map`` to coefficients.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.poly_tup_engine import apply_coeff_map
        sage: sq = lambda x : x**2
        sage: R.<x, y, z> = PolynomialRing(ZZ)
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup, _tup_to_poly
        sage: _tup_to_poly(apply_coeff_map(poly_to_tup(x + 2*y + 3*z), sq), parent=R)
        x + 4*y + 9*z
    """
    cdef ETuple exp
    cdef list new_tup = []
    for exp, coeff in eq_tup:
        new_tup.append((exp, coeff_map(coeff)))
    return tuple(new_tup)

# cpdef inline bint tup_fixes_sq(tuple eq_tup):
cdef inline bint tup_fixes_sq(tuple eq_tup) noexcept:
    r"""
    Determine if given equation fixes the square of a variable.

    An equation fixes the sq of a variable if it is of the form `a*x^2 + c`
    for *nonzero* constants `a`, `c`.
    """
    if len(eq_tup) != 2:
        return False
    # To access _attributes, we must cdef ETuple
    cdef ETuple lm = eq_tup[0][0]
    if lm._nonzero != 1 or lm._data[1] != 2:
        return False
    cdef ETuple tm = eq_tup[1][0]
    if tm._nonzero != 0:
        return False
    return True

######################
#   Simplification   #
######################

cdef dict subs_squares(dict eq_dict, KSHandler known_sq):
    r"""
    Substitute for known squares into a given polynomial.

    INPUT:

    - ``eq_dict`` -- dictionary of ``(ETuple, coeff)`` pairs representing
      a polynomial
    - ``known_sq`` -- dictionary of ``(int i, NumberFieldElement a)`` pairs
      such that `x_i^2 - a = 0`

    OUTPUT:

    A dictionary of ``(ETuple, coeff)`` pairs.
    """
    cdef dict subbed, new_e
    cdef ETuple exp
    cdef int idx, power
    subbed = dict()
    for exp, coeff in eq_dict.items():
        new_e = dict()
        for idx, power in exp.sparse_iter():
            if known_sq.contains(idx):
                coeff *= pow(known_sq.get(idx), power // 2)
                # New power is 1 if power is odd
                if power & True:
                    new_e[idx] = 1
            else:
                new_e[idx] = power
        exp = ETuple(new_e, len(exp))
        # If exponent tuple is already present in dictionary, coefficients are added
        if exp in subbed:
            subbed[exp] += coeff
        else:
            subbed[exp] = coeff
    return subbed

cdef dict remove_gcf(dict eq_dict, ETuple nonz):
    r"""
    Return a dictionary of ``(ETuple, coeff)`` pairs describing the
    polynomial ``eq / GCF(eq)``.

    The input ``nonz`` is an ``ETuple`` indicating the positions of
    variables known to be nonzero. The entries of ``nonz`` are assumed to
    be some relatively large number, like 100.
    """
    # Find common variables, filtered according to known nonzeros
    cdef ETuple common_powers, exp
    cdef NumberFieldElement_absolute c
    common_powers = nonz
    for exp, c in eq_dict.items():
        common_powers = common_powers.emin(exp)
    cdef dict ret = {}
    for exp, c in eq_dict.items():
        ret[exp.esub(common_powers)] = c
    return ret

cdef tuple to_monic(dict eq_dict, one):
    """
    Return tuple of pairs ``(ETuple, coeff)`` describing the monic polynomial
    associated to ``eq_dict``.

    Here, the leading coefficient is chosen according to the degree reverse
    lexicographic ordering (default for multivariate polynomial rings).
    """
    if not eq_dict:
        return ()
    cdef list ord_monoms = sorted(eq_dict, key=monom_sortkey)
    cdef ETuple lm = ord_monoms[-1]
    cdef NumberFieldElement_absolute lc = eq_dict[lm]
    if not lc:
        return ()
    cdef list ret = [(lm, one)]
    inv_lc = lc.inverse_of_unit()
    cdef int i, n
    n = len(ord_monoms)
    for i in range(n-1):
        ret.append((ord_monoms[n-2-i], inv_lc * eq_dict[ord_monoms[n-2-i]]))
    return tuple(ret)

cdef tuple reduce_poly_dict(dict eq_dict, ETuple nonz, KSHandler known_sq, NumberFieldElement_absolute one):
    """
    Return a tuple describing a monic polynomial with no known nonzero
    gcf and no known squares.
    """
    if not eq_dict:
        return ()
    cdef dict sq_rmvd = subs_squares(eq_dict, known_sq)
    cdef dict gcf_rmvd = remove_gcf(sq_rmvd, nonz)
    return to_monic(gcf_rmvd, one)

####################
#   Substitution   #
####################

cpdef dict compute_known_powers(max_degs, dict val_dict, one):
    """
    Pre-compute powers of known values for efficiency when preparing to
    substitute into a list of polynomials.

    INPUT:

    - ``max_deg`` -- an ``ETuple`` indicating the maximal degree of
      each variable
    - ``val_dict`` -- dictionary of ``(var_idx, poly_tup)`` key-value pairs
    - ``poly_tup`` -- tuple of ``(ETuple, coeff)`` pairs representing a
      multivariate polynomial

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.poly_tup_engine import compute_known_powers
        sage: R.<x, y, z> = PolynomialRing(QQ)
        sage: polys = [x**3 + 1, x**2*y + z**3, y**2 - 3*y]
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: known_val = { 0 : poly_to_tup(R(-1)), 2 : poly_to_tup(y**2) }
        sage: from sage.algebras.fusion_rings.poly_tup_engine import get_variables_degrees
        sage: max_deg = get_variables_degrees([poly_to_tup(p) for p in polys], 3)
        sage: compute_known_powers(max_deg, known_val, R.base_ring().one())
        {0: [(((0, 0, 0), 1),),
             (((0, 0, 0), -1),),
             (((0, 0, 0), 1),),
             (((0, 0, 0), -1),)],
        2: [(((0, 0, 0), 1),),
            (((0, 2, 0), 1),),
            (((0, 4, 0), 1),),
            (((0, 6, 0), 1),)]}
    """
    assert (max_degs and max(max_degs) <= 100) or True, "cannot substitute for degree larger than 100"
    cdef ETuple max_deg = ETuple(list(max_degs))
    max_deg = max_deg.emin(ETuple({idx: 100 for idx in val_dict}, len(max_deg)))
    cdef dict known_powers
    # Get polynomial unit as tuple to initialize list elements
    cdef tuple one_tup = ((max_deg._new(), one), )
    cdef int d, power, var_idx
    known_powers = {var_idx: [one_tup]*(d+1) for var_idx, d in max_deg.sparse_iter()}
    for var_idx, d in max_deg.sparse_iter():
        for power in range(d):
            known_powers[var_idx][power+1] = tup_mul(known_powers[var_idx][power], val_dict[var_idx])
    return known_powers

cdef dict subs(tuple poly_tup, dict known_powers, one):
    """
    Substitute given variables into a polynomial tuple.
    """
    cdef dict subbed = {}
    cdef ETuple exp, m, shifted_exp
    cdef int var_idx, power
    cdef tuple temp
    for exp, coeff in poly_tup:
        # Get polynomial unit as tuple
        temp = ((exp._new(), one), )
        for var_idx, power in exp.sparse_iter():
            if var_idx in known_powers:
                exp = exp.eadd_p(-power, var_idx)
                temp = tup_mul(temp, known_powers[var_idx][power])
        for m, c in temp:
            shifted_exp = exp.eadd(m)
            if shifted_exp in subbed:
                subbed[shifted_exp] += coeff * c
            else:
                subbed[shifted_exp] = coeff * c
    return subbed

cdef tuple tup_mul(tuple p1, tuple p2):
    r"""
    Multiplication of two polynomial tuples using schoolbook multiplication.
    """
    cdef dict prod = {}
    cdef ETuple xi, yj, shifted_exp
    for xi, ai in p1:
        for yj, bj in p2:
            shifted_exp = xi.eadd(yj)
            if shifted_exp in prod:
                prod[shifted_exp] += ai * bj
            else:
                prod[shifted_exp] = ai * bj
    return tuple(prod.items())

###############
#   Sorting   #
###############

cdef tuple monom_sortkey(ETuple exp):
    r"""
    Produce a sortkey for a monomial exponent with respect to degree
    reversed lexicographic ordering.
    """
    cdef int deg = exp.unweighted_degree()
    # for i in range(exp._nonzero):
    # exp._data[2*i+1] = -exp._data[2*i+1]
    cdef ETuple rev = exp.reversed().emul(-1)
    return (deg, rev)

cpdef tuple poly_tup_sortkey(tuple eq_tup):
    r"""
    Return the sortkey of a polynomial represented as a tuple of
    ``(ETuple, coeff)`` pairs with respect to the degree
    lexicographical term order.

    Using this key to sort polynomial tuples results in comparing polynomials
    term by term (we assume the tuple representation is sorted so that the
    leading term with respect to the degree reverse lexicographical order
    comes first). For each term, we first compare degrees, then the monomials
    themselves. Different polynomials can have the same sortkey.

    EXAMPLES::

     sage: F = CyclotomicField(20)
     sage: zeta20 = F.gen()
     sage: R.<x, y, z> = PolynomialRing(F)
     sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_tup_sortkey, poly_to_tup
     sage: p = (zeta20 + 1)*x^2 + (zeta20^3 + 6)*x*z + (zeta20^2 + 7*zeta20)*z^2 + (2/3*zeta20 + 1/4)*x + y
     sage: p1 = poly_to_tup(p); p1
     (((2, 0, 0), zeta20 + 1),
      ((1, 0, 1), zeta20^3 + 6),
      ((0, 0, 2), zeta20^2 + 7*zeta20),
      ((1, 0, 0), 2/3*zeta20 + 1/4),
      ((0, 1, 0), 1))
        sage: poly_tup_sortkey(p1)
        (2, 0, 2, 2, 0, 1, -2, 1, 2, -2, 2, 1, 0, 1, 1, -1, 1)
    """
    cdef ETuple exp
    cdef int i
    cdef list key = []
    for exp, c in eq_tup:
        # Compare by term degree
        key.append(exp.unweighted_degree())
        # Next compare by term w.r.t. lex order
        for i in range(exp._nonzero):
            # key.append(exp._length-1-exp._data[2*(nnz-i-1)])
            # key.append(-exp._data[2*(nnz-i-1)+1])
            key.append(-exp._data[2*i])
            key.append(exp._data[2*i+1])
    return tuple(key)



# sage_setup: distribution = sagemath-singular
# sage.doctest: needs sage.combinat sage.groups
"""
Fast Fusion Ring Methods for Computing Braid Group Representations
"""
# ****************************************************************************
#  Copyright (C) 2021 Guillermo Aboumrad <gh_willieab>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from ctypes import cast, py_object
cimport cython
from sage.algebras.fusion_rings.fast_parallel_fmats_methods cimport _fmat

from sage.rings.qqbar import QQbar

###############
#   Mappers   #
###############

cdef mid_sig_ij(fusion_ring, row, col, a, b):
    r"""
    Compute the (xi, yi), (xj, yj) entry of generator braiding the middle two
    strands in the tree b -> xi # yi -> (a # a) # (a # a), which results in
    a sum over j of trees b -> xj # yj -> (a # a) # (a # a)

    .. WARNING::

        This method assumes F-matrices are orthogonal.
    """
    # Pre-compute common parameters for efficiency
    _fvars = fusion_ring.get_fmatrix()._fvars
    _Nk_ij = fusion_ring.Nk_ij
    one = fusion_ring.one()

    xi, yi = row
    xj, yj = col
    entry = 0
    cdef list basis = list(fusion_ring.basis())
    for c in basis:
        for d in basis:
            # #Warning: We assume F-matrices are orthogonal!!! (using transpose for inverse)
            f1 = _fmat(_fvars, _Nk_ij, one, a, a, yi, b, xi, c)
            f2 = _fmat(_fvars, _Nk_ij, one, a, a, a, c, d, yi)
            f3 = _fmat(_fvars, _Nk_ij, one, a, a, a, c, d, yj)
            f4 = _fmat(_fvars, _Nk_ij, one, a, a, yj, b, xj, c)
            r = fusion_ring.r_matrix(a, a, d)
            entry += f1 * f2 * r * f3 * f4
    return entry

cdef odd_one_out_ij(fusion_ring, xi, xj, a, b):
    r"""
    Compute the `xi`, `xj` entry of the braid generator on the two right-most
    strands, corresponding to the tree b -> (xi # a) -> (a # a) # a, which
    results in a sum over j of trees b -> xj -> (a # a) # (a # a)

    .. WARNING::

        This method assumes F-matrices are orthogonal.
    """
    # Pre-compute common parameters for efficiency
    _fvars = fusion_ring.get_fmatrix()._fvars
    _Nk_ij = fusion_ring.Nk_ij
    one = fusion_ring.one()

    entry = 0
    for c in fusion_ring.basis():
        # #Warning: We assume F-matrices are orthogonal!!! (using transpose for inverse)
        f1 = _fmat(_fvars, _Nk_ij, one, a, a, a, b, xi, c)
        f2 = _fmat(_fvars, _Nk_ij, one, a, a, a, b, xj, c)
        r = fusion_ring.r_matrix(a, a, c)
        entry += f1 * r * f2
    return entry

# Cache methods (manually for cdef methods)
cdef odd_one_out_ij_cache = dict()
cdef mid_sig_ij_cache = dict()

cdef cached_mid_sig_ij(fusion_ring, row, col, a, b):
    r"""
    Cached version of :meth:`mid_sig_ij`.
    """
    if (row, col, a, b) in mid_sig_ij_cache:
        return mid_sig_ij_cache[row, col, a, b]
    entry = mid_sig_ij(fusion_ring, row, col, a, b)
    mid_sig_ij_cache[row, col, a, b] = entry
    return entry

cdef cached_odd_one_out_ij(fusion_ring, xi, xj, a, b):
    r"""
    Cached version of :meth:`odd_one_out_ij`.
    """
    if (xi, xj, a, b) in odd_one_out_ij_cache:
        return odd_one_out_ij_cache[xi, xj, a, b]
    entry = odd_one_out_ij(fusion_ring, xi, xj, a, b)
    odd_one_out_ij_cache[xi, xj, a, b] = entry
    return entry


@cython.nonecheck(False)
@cython.cdivision(True)
cdef sig_2k(fusion_ring, tuple args):
    r"""
    Compute entries of the `2k`-th braid generator
    """
    # Pre-compute common parameters for efficiency
    _fvars = fusion_ring.get_fmatrix()._fvars
    _Nk_ij = fusion_ring.Nk_ij
    one = fusion_ring.one()

    cdef int child_id, n_proc
    child_id, n_proc, fn_args = args
    k, a, b, n_strands = fn_args
    cdef int ctr = -1
    cdef list worker_results = list()
    # Get computational basis
    cdef list comp_basis = fusion_ring.get_computational_basis(a, b, n_strands)
    cdef dict basis_dict = {elt: i for i, elt in enumerate(comp_basis)}
    cdef int dim = len(comp_basis)
    cdef set coords = set()
    cdef int i
    # Avoid pickling cyclotomic field element objects
    cdef bint must_flatten_coeff = fusion_ring.fvars_field() != QQbar
    cdef list basis = list(fusion_ring.basis())
    for i in range(dim):
        for f in basis:
            for e in basis:
                for q in basis:
                    # Distribute work amongst processes
                    ctr += 1
                    if ctr % n_proc != child_id:
                        continue

                    # Compute appropriate possible nonzero row index
                    nnz_pos = list(comp_basis[i])
                    nnz_pos[k-1] = f
                    nnz_pos[k] = e
                    # Handle the special case k = 1
                    if k > 1:
                        nnz_pos[n_strands//2+k-2] = q
                    nnz_pos = tuple(nnz_pos)

                    # Skip repeated entries when k = 1
                    if nnz_pos in comp_basis and (basis_dict[nnz_pos], i) not in coords:
                        m, l = comp_basis[i][:n_strands//2], comp_basis[i][n_strands//2:]
                        # A few special cases
                        top_left = m[0]
                        if k >= 3:
                            top_left = l[k-3]
                        root = b
                        if k - 1 < len(l):
                            root = l[k-1]

                        # Handle the special case k = 1
                        if k == 1:
                            entry = cached_mid_sig_ij(fusion_ring, m[:2], (f, e), a, root)

                            # Avoid pickling cyclotomic field element objects
                            if must_flatten_coeff:
                                entry = entry.list()

                            worker_results.append(((basis_dict[nnz_pos], i), entry))
                            coords.add((basis_dict[nnz_pos], i))
                            continue

                        entry = 0
                        for p in fusion_ring.basis():
                            f1 = _fmat(_fvars, _Nk_ij, one, top_left, m[k-1], m[k], root, l[k-2], p)
                            f2 = _fmat(_fvars, _Nk_ij, one, top_left, f, e, root, q, p)
                            entry += f1 * cached_mid_sig_ij(fusion_ring, (m[k-1], m[k]), (f, e), a, p) * f2

                        # Avoid pickling cyclotomic field element objects
                        if must_flatten_coeff:
                            entry = entry.list()

                        worker_results.append(((basis_dict[nnz_pos], i), entry))
    return worker_results


@cython.nonecheck(False)
@cython.cdivision(True)
cdef odd_one_out(fusion_ring, tuple args):
    r"""
    Compute entries of the rightmost braid generator, in case we have an
    odd number of strands.
    """
    # Pre-compute common parameters for efficiency
    _fvars = fusion_ring.get_fmatrix()._fvars
    _Nk_ij = fusion_ring.Nk_ij
    one = fusion_ring.one()

    cdef list worker_results = []
    cdef list nnz_pos_temp
    cdef tuple nnz_pos
    cdef int child_id, n_proc, i
    child_id, n_proc, fn_args = args
    a, b, n_strands = fn_args
    cdef int ctr = -1
    # Get computational basis
    cdef list comp_basis = fusion_ring.get_computational_basis(a, b, n_strands)
    cdef dict basis_dict = {elt: i for i, elt in enumerate(comp_basis)}
    cdef int dim = len(comp_basis)

    # Avoid pickling cyclotomic field element objects
    cdef bint must_flatten_coeff = fusion_ring.fvars_field() != QQbar

    cdef list basis = list(fusion_ring.basis())
    for i in range(dim):
        for f in basis:
            for q in basis:
                # Distribute work amongst processes
                ctr += 1
                if ctr % n_proc != child_id:
                    continue

                # Compute appropriate possible nonzero row index
                nnz_pos_temp = list(comp_basis[i])
                nnz_pos_temp[n_strands//2-1] = f
                # Handle small special case
                if n_strands > 3:
                    nnz_pos_temp[-1] = q
                nnz_pos = tuple(nnz_pos_temp)

                if nnz_pos in comp_basis:
                    m, l = comp_basis[i][:n_strands//2], comp_basis[i][n_strands//2:]

                    # Handle a couple of small special cases
                    if n_strands == 3:
                        entry = cached_odd_one_out_ij(fusion_ring, m[-1], f, a, b)

                        # Avoid pickling cyclotomic field element objects
                        if must_flatten_coeff:
                            entry = entry.list()

                        worker_results.append(((basis_dict[nnz_pos], i), entry))
                        continue
                    top_left = m[0]
                    if n_strands > 5:
                        top_left = l[-2]
                    root = b

                    # Compute relevant entry
                    entry = 0
                    for p in basis:
                        f1 = _fmat(_fvars, _Nk_ij, one, top_left, m[-1], a, root, l[-1], p)
                        f2 = _fmat(_fvars, _Nk_ij, one, top_left, f, a, root, q, p)
                        entry += f1 * cached_odd_one_out_ij(fusion_ring, m[-1], f, a, p) * f2

                    # Avoid pickling cyclotomic field element objects
                    if must_flatten_coeff:
                        entry = entry.list()

                    worker_results.append(((basis_dict[nnz_pos], i), entry))
    return worker_results

##############################
#   Parallel code executor   #
##############################

# Hard-coded module __dict__-style attribute with visible cdef methods
cdef dict mappers = {
    "sig_2k": sig_2k,
    "odd_one_out": odd_one_out
}

cpdef executor(tuple params):
    r"""
    Execute a function registered in this module's ``mappers``
    in a worker process, and supply the ``FusionRing`` parameter by
    constructing a reference to the FMatrix object in the worker's memory
    address space from its ``id``.

    .. NOTE::

        When the parent process is forked, each worker gets a copy of
        every  global variable. The virtual memory address of object `X` in
        the parent process equals the *virtual* memory address of the copy of
        object `X` in each worker, so we may construct references to forked
        copies of `X`.

    TESTS::

        sage: from sage.algebras.fusion_rings.fast_parallel_fusion_ring_braid_repn import executor
        sage: FR = FusionRing("A1", 4)
        sage: FR.fusion_labels(['idd', 'one', 'two', 'three', 'four'], inject_variables=True)
        sage: FR.get_fmatrix().find_orthogonal_solution(verbose=False)    # long time
        sage: params = (('sig_2k', id(FR)), (0, 1, (1, one, one, 5)))    # long time
        sage: len(executor(params)) == 13                         # long time
        True
        sage: from sage.algebras.fusion_rings.fast_parallel_fusion_ring_braid_repn import executor
        sage: FR = FusionRing("A1", 2)
        sage: FR.fusion_labels("a", inject_variables=True)
        sage: FR.get_fmatrix().find_orthogonal_solution(verbose=False)
        sage: params = (('odd_one_out', id(FR)), (0, 1, (a2, a2, 5)))
        sage: len(executor(params)) == 1
        True
    """
    (fn_name, fr_id), args = params
    # Construct a reference to global FMatrix object in this worker's memory
    fusion_ring_obj = cast(fr_id, py_object).value
    # Bind module method to FMatrix object in worker process, and call the method
    return mappers[fn_name](fusion_ring_obj, args)

######################################
#   Pickling circumvention helpers   #
######################################

cpdef _unflatten_entries(fusion_ring, list entries):
    r"""
    Restore cyclotomic coefficient object from its tuple of rational
    coefficients representation.

    Used to circumvent pickling issue introduced by PARI settings
    in :issue:`30537`.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.fast_parallel_fusion_ring_braid_repn import _unflatten_entries
        sage: fr = FusionRing("B2", 2)
        sage: F = fr.field()
        sage: coeff = [F.random_element() for i in range(2)]
        sage: entries = [((0, 0), coeff[0].list()), ((0, 1), coeff[1].list())]
        sage: _unflatten_entries(fr, entries)
        sage: all(cyc_elt_obj == c for (coord, cyc_elt_obj), c in zip(entries, coeff))
        True
    """
    F = fusion_ring.fvars_field()
    if F != QQbar:
        for i, (coord, entry) in enumerate(entries):
            entries[i] = (coord, F(entry))



# sage_setup: distribution = sagemath-singular
# sage.doctest: needs sage.combinat sage.groups
r"""
Shared Memory Managers for F-Symbol Attributes

This module provides an implementation for shared dictionary like
state attributes required by the orthogonal F-matrix solver.

Currently, the attributes only work when the base field of the
:class:`FMatrix` factory is a cyclotomic field.
"""

# ****************************************************************************
#  Copyright (C) 2021 Guillermo Aboumrad <gh_willieab>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

cimport cython
cimport numpy as np
from cysignals.memory cimport sig_malloc
from multiprocessing import shared_memory
from sage.algebras.fusion_rings.poly_tup_engine cimport poly_to_tup, tup_fixes_sq, _flatten_coeffs
from sage.rings.integer cimport Integer
from sage.rings.rational cimport Rational
from sage.rings.polynomial.multi_polynomial cimport MPolynomial_libsingular
from sage.rings.polynomial.polydict cimport ETuple

import numpy as np
from os import getpid

cdef class KSHandler:
    r"""
    A shared memory backed dict-like structure to manage the
    ``_ks`` attribute of an F-matrix.

    This structure implements a representation of the known squares dictionary
    using a structured NumPy array backed by a contiguous shared memory
    object.

    The structure mimics a dictionary of ``(idx, known_sq)`` pairs. Each
    integer index corresponds to a variable and each ``known_sq`` is an
    element of the F-matrix factory's base cyclotomic field.

    Each cyclotomic coefficient is stored as a list of numerators and a
    list of denominators representing the rational coefficients. The
    structured array also maintains ``known`` attribute that indicates
    whether the structure contains an entry corresponding to the given index.

    The parent process should construct this object without a
    ``name`` attribute. Children processes use the ``name`` attribute,
    accessed via ``self.shm.name`` to attach to the shared memory block.

    INPUT:

    - ``n_slots`` -- the total number of F-symbols
    - ``field`` -- F-matrix's base cyclotomic field
    - ``use_mp`` -- boolean indicating whether to construct a shared
      memory block to back ``self``
    - ``init_data`` -- dictionary or :class:`KSHandler` object containing
      known squares for initialization, e.g., from a solver checkpoint
    - ``name`` -- the name of a shared memory object (used by child processes
        for attaching)

    .. NOTE::

        To properly dispose of shared memory resources,
        ``self.shm.unlink()`` must be called before exiting.

    .. WARNING::

        This structure *cannot* modify an entry that
        has already been set.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.shm_managers import KSHandler
        sage: # Create shared data structure
        sage: f = FusionRing("A1", 2).get_fmatrix(inject_variables=True, new=True)
        creating variables fx1..fx14
        Defining fx0, fx1, fx2, fx3, fx4, fx5, fx6, fx7, fx8, fx9, fx10, fx11, fx12, fx13
        sage: n = f._poly_ring.ngens()
        sage: f.start_worker_pool()
        sage: ks = KSHandler(n, f._field, use_mp=True)
        sage: # In the same shell or in a different shell, attach to fvars
        sage: name = ks.shm.name
        sage: ks2 = KSHandler(n, f._field, name=name, use_mp=True)
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: eqns = [fx1**2 - 4, fx3**2 + f._field.gen()**4 - 1/19*f._field.gen()**2]
        sage: ks.update([poly_to_tup(p) for p in eqns])
        sage: for idx, sq in ks.items():
        ....:     print("Index: {}, square: {}".format(idx, sq))
        ....:
        Index: 1, square: 4
        Index: 3, square: -zeta32^4 + 1/19*zeta32^2
        sage: ks.shm.unlink()
        sage: f.shutdown_worker_pool()
    """
    def __init__(self, n_slots, field, use_mp=False, init_data=None, name=None):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.algebras.fusion_rings.shm_managers import KSHandler
            sage: # Create shared data structure
            sage: f = FusionRing("A1", 2).get_fmatrix(inject_variables=True, new=True)
            creating variables fx1..fx14
            Defining fx0, fx1, fx2, fx3, fx4, fx5, fx6, fx7, fx8, fx9, fx10, fx11, fx12, fx13
            sage: n = f._poly_ring.ngens()
            sage: f.start_worker_pool()
            sage: ks = KSHandler(n, f._field, use_mp=True)
            sage: TestSuite(ks).run()
            sage: ks.shm.unlink()
            sage: f.shutdown_worker_pool()
        """
        cdef int n, d
        if init_data is None:
            init_data = {}
        self.field = field
        n = n_slots
        d = self.field.degree()
        ks_t = np.dtype([
            ('known', 'bool', (1, )),
            ('nums', 'i8', (d, )),
            ('denoms', 'u8', (d, ))
        ])
        self.obj_cache = [None]*n
        if use_mp:
            if name is None:
                self.shm = shared_memory.SharedMemory(create=True, size=n*ks_t.itemsize)
            else:
                self.shm = shared_memory.SharedMemory(name=name)
            self.ks_dat = np.ndarray((n, ), dtype=ks_t, buffer=self.shm.buf)
        else:
            self.ks_dat = np.ndarray((n, ), dtype=ks_t)
        if name is None:
            self.ks_dat['known'] = np.zeros((n, 1), dtype='bool')
            self.ks_dat['nums'] = np.zeros((n, d), dtype='i8')
            self.ks_dat['denoms'] = np.ones((n, d), dtype='u8')
        # Populate initializer data
        for idx, sq in init_data.items():
            self.setitem(idx, sq)

    @cython.nonecheck(False)
    @cython.wraparound(False)
    @cython.boundscheck(False)
    cdef NumberFieldElement_absolute get(self, int idx):
        r"""
        Retrieve the known square corresponding to the given index,
        if it exists.
        """
        if not self.ks_dat['known'][idx]:
            raise KeyError('index {} does not correspond to a known square'.format(idx))
        if self.obj_cache[idx] is not None:
            return self.obj_cache[idx]
        cdef int d
        cdef list rat
        cdef Py_ssize_t i
        cdef np.ndarray[np.int64_t, ndim=1] nums = self.ks_dat['nums'][idx]
        # cdef np.int64_t[::1] num_view = nums
        cdef np.ndarray[np.uint64_t, ndim=1] denoms = self.ks_dat['denoms'][idx]
        # cdef np.uint64_t[::1] denom_view = denoms
        cdef np.int64_t num
        cdef np.uint64_t denom
        cdef NumberFieldElement_absolute cyc_coeff
        cdef Rational quo
        d = self.field.degree()
        rat = list()
        for i in range(d):
            num = nums[i]
            denom = denoms[i]
            quo = Integer(num) / Integer(denom)
            rat.append(quo)
        cyc_coeff = self.field(rat)
        self.obj_cache[idx] = cyc_coeff
        return cyc_coeff

    cpdef update(self, list eqns):
        r"""
        Update ``self``'s ``shared_memory``-backed dictionary of known
        squares. Keys are variable indices and corresponding values
        are the squares.

        EXAMPLES::

            sage: f = FusionRing("B5", 1).get_fmatrix()
            sage: f._reset_solver_state()
            sage: for idx, sq in f._ks.items():
            ....:     k
            ....:
            sage: f.get_orthogonality_constraints()
            [fx0^2 - 1,
             fx1^2 - 1,
             fx2^2 - 1,
             fx3^2 - 1,
             fx4^2 - 1,
             fx5^2 - 1,
             fx6^2 - 1,
             fx7^2 - 1,
             fx8^2 - 1,
             fx9^2 - 1,
             fx10^2 + fx12^2 - 1,
             fx10*fx11 + fx12*fx13,
             fx10*fx11 + fx12*fx13,
             fx11^2 + fx13^2 - 1]
             sage: f.get_orthogonality_constraints(output=False)
             sage: f._ks.update(f.ideal_basis)
             sage: for idx, sq in f._ks.items():
             ....:     print(idx, "-->", sq)
             ....:
             0 --> 1
             1 --> 1
             2 --> 1
             3 --> 1
             4 --> 1
             5 --> 1
             6 --> 1
             7 --> 1
             8 --> 1
             9 --> 1

        .. WARNING::

            This method assumes every polynomial in ``eqns`` is *monic*.
        """
        cdef ETuple lm
        cdef list rhs
        cdef Py_ssize_t i, idx
        cdef tuple eq_tup
        for i in range(len(eqns)):
            eq_tup = eqns[i]
            if tup_fixes_sq(eq_tup):
                rhs = [-v for v in eq_tup[-1][1]]
                # eq_tup is guaranteed univariate, so we extract variable idx from lm
                lm = eq_tup[0][0]
                idx = lm._data[0]
                try:
                    self.setitem(idx, rhs)
                except OverflowError:
                    print("KS overflowed on index {} with value {}".format(idx, self.field(rhs)))

    @cython.nonecheck(False)
    @cython.wraparound(False)
    @cython.infer_types(False)
    cdef setitem(self, int idx, rhs):
        """
        Create an entry corresponding to the given index.

        The ``rhs`` parameter may be a cyclotomic coefficient or its
        list/tuple representation.
        """
        cdef Py_ssize_t i
        cdef np.ndarray[np.int64_t, ndim=1] nums = self.ks_dat['nums'][idx]
        cdef np.ndarray[np.uint64_t, ndim=1] denoms = self.ks_dat['denoms'][idx]
        cdef np.int64_t num
        cdef np.uint64_t denom
        cdef Rational quo
        self.ks_dat['known'][idx] = True
        if not isinstance(rhs, list):
            rhs = rhs._coefficients()
        for i in range(len(rhs)):
            quo = rhs[i]
            num = quo.numerator()
            denom = quo.denominator()
            if num > 2**32:
                print("WARNING: Large num encountered in KS", num)
            if denom > 2**32:
                print("WARNING: Large denom encountered in KS", denom)
            nums[i] = num
            denoms[i] = denom

    cdef bint contains(self, int idx) noexcept:
        r"""
        Determine whether ``self`` contains entry corresponding to given
        ``idx``.
        """
        return self.ks_dat[idx]['known']

    def __eq__(self, KSHandler other):
        r"""
        Test for equality.

        TESTS::

            sage: f = FusionRing("C2", 2).get_fmatrix()
            sage: f._reset_solver_state()
            sage: f.get_orthogonality_constraints(output=False)
            sage: from sage.algebras.fusion_rings.shm_managers import KSHandler
            sage: n = f._poly_ring.ngens()
            sage: f.start_worker_pool()
            sage: ks = KSHandler(n, f._field, use_mp=True, init_data=f._ks)
            sage: # In the same shell or in a different one, attach to shared memory handler
            sage: name = ks.shm.name
            sage: k2 = KSHandler(n, f._field, name=name, use_mp=True)
            sage: ks == k2
            True
            sage: ks.shm.unlink()
            sage: f.shutdown_worker_pool()
        """
        return all(other.get(idx) == sq for idx, sq in self.items())

    def __reduce__(self):
        r"""
        Provide pickling / unpickling support for ``self``.

        TESTS::

            sage: f = FusionRing("A3", 1).get_fmatrix()
            sage: f._reset_solver_state()
            sage: loads(dumps(f._ks)) == f._ks
            True
            sage: f.find_orthogonal_solution(verbose=False)    # long time
            sage: loads(dumps(f._ks)) == f._ks
            True
        """
        d = {i: sq for i, sq in self.items()}
        return make_KSHandler, (self.ks_dat.size, self.field, d)

    def items(self):
        r"""
        Iterate through existing entries using Python dict-style syntax.

        EXAMPLES::

            sage: f = FusionRing("A3", 1).get_fmatrix()
            sage: f._reset_solver_state()
            sage: f.get_orthogonality_constraints(output=False)
            sage: f._ks.update(f.ideal_basis)
            sage: for idx, sq in f._ks.items():
            ....:     print("Index: {}, sq: {}".format(idx, sq))
            ....:
            Index: 0, sq: 1
            Index: 1, sq: 1
            Index: 2, sq: 1
            Index: 3, sq: 1
            Index: 4, sq: 1
            ...
            Index: 25, sq: 1
            Index: 26, sq: 1
        """
        cdef Py_ssize_t i
        for i in range(self.ks_dat.size):
            if self.ks_dat['known'][i]:
                yield i, self.get(i)


def make_KSHandler(n_slots, field, init_data):
    r"""
    Provide pickling / unpickling support for :class:`KSHandler`.

    TESTS::

        sage: f = FusionRing("B4", 1).get_fmatrix()
        sage: f._reset_solver_state()
        sage: loads(dumps(f._ks)) == f._ks                 # indirect doctest
        True
        sage: f.find_orthogonal_solution(verbose=False)    # long time
        sage: loads(dumps(f._ks)) == f._ks                 # indirect doctest
        True
    """
    return KSHandler(n_slots, field, init_data=init_data)


cdef class FvarsHandler:
    r"""
    A shared memory backed dict-like structure to manage the
    ``_fvars`` attribute of an F-matrix.

    This structure implements a representation of the F-symbols dictionary
    using a structured NumPy array backed by a contiguous shared memory
    object.

    The monomial data is stored in the ``exp_data`` structure. Monomial
    exponent data is stored contiguously and ``ticks`` are used to
    indicate different monomials.

    Coefficient data is stored in the ``coeff_nums`` and ``coeff_denom``
    arrays. The ``coeff_denom`` array stores the value
    ``d = coeff.denominator()`` for each cyclotomic coefficient. The
    ``coeff_nums`` array stores the values
    ``c.numerator() * d for c in coeff._coefficients()``, the abridged
    list representation of the cyclotomic coefficient ``coeff``.

    Each entry also has a boolean ``modified`` attribute, indicating
    whether it has been modified by the parent process. Entry retrieval
    is cached in each process, so each process must check whether
    entries have been modified before attempting retrieval.

    The parent process should construct this object without a
    ``name`` attribute. Children processes use the ``name`` attribute,
    accessed via ``self.shm.name`` to attach to the shared memory block.

    INPUT:

    - ``n_slots`` -- number of generators of the underlying polynomial ring
    - ``field`` -- base field for polynomial ring
    - ``idx_to_sextuple`` -- map relating a single integer index to a sextuple
      of ``FusionRing`` elements
    - ``init_data`` -- dictionary or :class:`FvarsHandler` object containing
      known squares for initialization, e.g., from a solver checkpoint
    - ``use_mp`` -- integer indicating the number of child processes
      used for multiprocessing; if running serially, use 0
    - ``pids_name`` -- the name of a ``ShareableList`` containing the
      process ``pid``'s for every process in the pool (including the
      parent process)
    - ``name`` -- the name of a shared memory object
      (used by child processes for attaching)
    - ``max_terms`` -- maximum number of terms in each entry; since
      we use contiguous C-style memory blocks, the size of the block
      must be known in advance
    - ``n_bytes`` -- the number of bytes that should be allocated for
      each numerator and each denominator stored by the structure

    .. NOTE::

        To properly dispose of shared memory resources,
        ``self.shm.unlink()`` must be called before exiting.

    .. NOTE::

        If you ever encounter an :exc:`OverflowError` when running the
        :meth:`FMatrix.find_orthogonal_solution` solver, consider
        increasing the parameter ``n_bytes``.

    .. WARNING::

        The current data structure supports up to `2^16` entries,
        with each monomial in each entry having at most 254
        nonzero terms. On average, each of the ``max_terms`` monomials
        can have at most 30 terms.

    EXAMPLES::

        sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
        sage: # Create shared data structure
        sage: f = FusionRing("A2", 1).get_fmatrix(inject_variables=True, new=True)
        creating variables fx1..fx8
        Defining fx0, fx1, fx2, fx3, fx4, fx5, fx6, fx7
        sage: f.start_worker_pool()
        sage: n_proc = f.pool._processes
        sage: pids_name = f._pid_list.shm.name
        sage: fvars = FvarsHandler(8, f._field, f._idx_to_sextuple, use_mp=n_proc, pids_name=pids_name)
        sage: # In the same shell or in a different shell, attach to fvars
        sage: name = fvars.shm.name
        sage: fvars2 = FvarsHandler(8, f._field, f._idx_to_sextuple, name=name , use_mp=n_proc, pids_name=pids_name)
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: rhs = tuple((exp, tuple(c._coefficients())) for exp, c in poly_to_tup(fx5**5))
        sage: fvars[f2, f1, f2, f2, f0, f0] = rhs
        sage: f._tup_to_fpoly(fvars2[f2, f1, f2, f2, f0, f0])
        fx5^5
        sage: fvars.shm.unlink()
        sage: f.shutdown_worker_pool()
    """
    def __init__(self, n_slots, field, idx_to_sextuple, init_data=None,
                 use_mp=0,
                 pids_name=None, name=None, max_terms=20, n_bytes=32):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
            sage: # Create shared data structure
            sage: f = FusionRing("A2", 1).get_fmatrix(inject_variables=True, new=True)
            creating variables fx1..fx8
            Defining fx0, fx1, fx2, fx3, fx4, fx5, fx6, fx7
            sage: f.start_worker_pool()
            sage: n_proc = f.pool._processes
            sage: pids_name = f._pid_list.shm.name
            sage: fvars = FvarsHandler(8, f._field, f._idx_to_sextuple, use_mp=n_proc, pids_name=pids_name)
            sage: TestSuite(fvars).run(skip='_test_pickling')
            sage: fvars.shm.unlink()
            sage: f.shutdown_worker_pool()
        """
        self.field = field
        self.obj_cache = dict()
        cdef int d = self.field.degree()
        self.bytes = n_bytes
        cdef int slots = self.bytes // 8
        cdef int n_proc = use_mp + 1
        if init_data is None:
            init_data = {}
        self.fvars_t = np.dtype([
            ('modified', np.int8, (n_proc, )),
            ('ticks', 'u1', (max_terms, )),
            ('exp_data', 'u2', (max_terms*30, )),
            ('coeff_nums', np.int64, (max_terms, d, slots)),
            ('coeff_denom', np.uint64, (max_terms, d, slots))
        ])
        self.sext_to_idx = {s: i for i, s in idx_to_sextuple.items()}
        self.ngens = n_slots
        if use_mp:
            if name is None:
                self.shm = shared_memory.SharedMemory(create=True, size=self.ngens*self.fvars_t.itemsize)
            else:
                self.shm = shared_memory.SharedMemory(name=name)
            self.fvars = np.ndarray((self.ngens, ), dtype=self.fvars_t, buffer=self.shm.buf)
            self.pid_list = shared_memory.ShareableList(name=pids_name)
            self.child_id = -1
        else:
            self.fvars = np.ndarray((self.ngens, ), dtype=self.fvars_t)
            self.child_id = 0
        # Populate with initialization data
        for sextuple, fvar in init_data.items():
            if isinstance(fvar, MPolynomial_libsingular):
                fvar = _flatten_coeffs(poly_to_tup(fvar))
            if isinstance(fvar, NumberFieldElement_absolute):
                fvar = ((ETuple({}, self.ngens), tuple(fvar._coefficients())), )
            if isinstance(fvar, tuple):
                transformed = list()
                for exp, c in fvar:
                    if isinstance(c, NumberFieldElement_absolute):
                        transformed.append((exp, tuple(c._coefficients())))
                if transformed:
                    fvar = tuple(transformed)
            self[sextuple] = fvar

    @cython.nonecheck(False)
    @cython.wraparound(False)
    @cython.boundscheck(False)
    def __getitem__(self, sextuple):
        r"""
        Retrieve a record from the shared memory data structure by
        unflattening its representation and constructing relevant Python
        objects.

        This method returns a tuple of ``(ETuple, coeff)`` pairs,
        where ``coeff`` is an element of ``self.field``.

        EXAMPLES::

            sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
            sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
            sage: f = FusionRing("B7", 1).get_fmatrix(inject_variables=True, new=True)
            creating variables fx1..fx14
            Defining fx0, fx1, fx2, fx3, fx4, fx5, fx6, fx7, fx8, fx9, fx10, fx11, fx12, fx13
            sage: f.start_worker_pool()
            sage: n_proc = f.pool._processes
            sage: pids_name = f._pid_list.shm.name
            sage: fvars = FvarsHandler(14, f._field, f._idx_to_sextuple, use_mp=n_proc, pids_name=pids_name)
            sage: rhs = tuple((exp, tuple(c._coefficients()))
            ....:             for exp, c in poly_to_tup(1/8*fx0**15 - 23/79*fx2*fx13**3 - 799/2881*fx1*fx2**5*fx10))
            sage: fvars[(f1, f2, f1, f2, f2, f2)] = rhs
            sage: rhs = tuple((exp, tuple(c._coefficients())) for exp, c in poly_to_tup(f._poly_ring.zero()))
            sage: fvars[f2, f2, f2, f2, f0, f0] = rhs
            sage: rhs = tuple((exp, tuple(c._coefficients())) for exp, c in poly_to_tup(-1/19*f._poly_ring.one()))
            sage: fvars[f2, f1, f2, f1, f2, f2] = rhs
            sage: s, t, r = (f1, f2, f1, f2, f2, f2), (f2, f2, f2, f2, f0, f0), (f2, f1, f2, f1, f2, f2)
            sage: f._tup_to_fpoly(fvars[s]) == 1/8*fx0**15 - 23/79*fx2*fx13**3 - 799/2881*fx1*fx2**5*fx10
            True
            sage: f._tup_to_fpoly(fvars[t]) == 0
            True
            sage: f._tup_to_fpoly(fvars[r]) == -1/19
            True
            sage: fvars.shm.unlink()
            sage: f.shutdown_worker_pool()

        .. NOTE::

            This method implements caching. Only the parent process is allowed
            to modify the shared fvars structure. Each process builds its own
            cache, so each process must update its cache before retrieving a
            modified entry, tagged via its ``modified`` property.
        """
        if sextuple not in self.sext_to_idx:
            raise KeyError('invalid sextuple {}'.format(sextuple))
        cdef Py_ssize_t idx = self.sext_to_idx[sextuple]
        # Each process builds its own cache, so each process must know
        # whether the entry it wants to retrieve has been modified.
        # Each process needs to know where to look, so we use an index
        # every process agrees on. The pid_list[0] belongs to the main process.
        if self.child_id < 0:
            self.child_id = self.pid_list.index(getpid())
        if idx in self.obj_cache:
            if self.fvars['modified'][idx, self.child_id]:
                del self.obj_cache[idx]
            else:
                return self.obj_cache[idx]
        cdef ETuple e, exp
        cdef int count, nnz
        cdef Integer num
        cdef list poly_tup, rats
        cdef NumberFieldElement_absolute cyc_coeff
        cdef Py_ssize_t cum, i, j, k
        cdef Rational quo
        cdef tuple ret
        # Define memory views to reduce Python overhead and ensure correct typing
        cdef np.ndarray[np.uint8_t, ndim=1] ticks = self.fvars['ticks'][idx]
        cdef np.ndarray[np.uint16_t, ndim=1] exp_data = self.fvars['exp_data'][idx]
        cdef np.ndarray[np.int64_t, ndim=3] nums = self.fvars['coeff_nums'][idx]
        cdef np.ndarray[np.uint64_t, ndim=3] denoms = self.fvars['coeff_denom'][idx]
        cdef np.ndarray[np.int8_t, ndim=1] modified = self.fvars['modified'][idx]
        e = ETuple({}, self.ngens)
        poly_tup = list()
        cum = 0
        count = np.count_nonzero(ticks)
        for i in range(count):
            # Construct new ETuple for each monomial
            exp = e._new()
            # Handle constant coeff
            nnz = ticks[i] if ticks[i] < 255 else 0
            exp._nonzero = nnz
            if nnz:
                exp._data = <int*>sig_malloc(sizeof(int)*nnz*2)
                for j in range(2*nnz):
                    exp._data[j] = <int>exp_data[cum]
                    cum += 1

            # Construct cyclotomic field coefficient
            rats = list()
            for k in range(self.field.degree()):
                num = Integer(list(nums[i, k]), 2**63)
                denom = Integer(list(denoms[i, k]), 2**64)
                quo = num / denom
                rats.append(quo)
            cyc_coeff = self.field(rats)
            poly_tup.append((exp, cyc_coeff))
        ret = tuple(poly_tup)
        # Cache object and reset modified
        self.obj_cache[idx] = ret
        modified[self.child_id] = 0
        return ret

    @cython.nonecheck(False)
    @cython.wraparound(False)
    def __setitem__(self, sextuple, fvar):
        r"""
        Given a sextuple of labels and a tuple of ``(ETuple, cyc_coeff)`` pairs,
        create or overwrite an entry in the shared data structure
        corresponding to the given sextuple.

        EXAMPLES::

            sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
            sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
            sage: f = FusionRing("A3", 1).get_fmatrix(inject_variables=True, new=True)
            creating variables fx1..fx27
            Defining fx0, ..., fx26
            sage: f.start_worker_pool()
            sage: n_proc = f.pool._processes
            sage: pids_name = f._pid_list.shm.name
            sage: fvars = FvarsHandler(27, f._field, f._idx_to_sextuple, use_mp=n_proc, pids_name=pids_name)
            sage: rhs = tuple((exp, tuple(c._coefficients()))
            ....:             for exp, c in poly_to_tup(1/8*fx0**15 - 23/79*fx2*fx21**3 - 799/2881*fx1*fx2**5*fx10))
            sage: fvars[(f3, f2, f1, f2, f1, f3)] = rhs
            sage: rhs = tuple((exp, tuple(c._coefficients())) for exp, c in poly_to_tup(f._poly_ring.zero()))
            sage: fvars[f3, f2, f3, f0, f1, f1] = rhs
            sage: rhs = tuple((exp, tuple(c._coefficients())) for exp, c in poly_to_tup(-1/19*f._poly_ring.one()))
            sage: fvars[f3, f3, f3, f1, f2, f2] = rhs
            sage: s, t, r = (f3, f2, f1, f2, f1, f3), (f3, f2, f3, f0, f1, f1), (f3, f3, f3, f1, f2, f2)
            sage: f._tup_to_fpoly(fvars[s]) == 1/8*fx0**15 - 23/79*fx2*fx21**3 - 799/2881*fx1*fx2**5*fx10
            True
            sage: f._tup_to_fpoly(fvars[t]) == 0
            True
            sage: f._tup_to_fpoly(fvars[r]) == -1/19
            True
            sage: fvars.shm.unlink()
            sage: f.shutdown_worker_pool()
        """
        cdef ETuple exp
        cdef Integer num, denom
        cdef tuple coeff_tup
        cdef Py_ssize_t cum, i, idx, j, k, t
        cdef Rational r
        idx = self.sext_to_idx[sextuple]
        # Clear entry before inserting
        self.fvars[idx] = np.zeros((1, ), dtype=self.fvars_t)
        # Define memory views to reduce Python overhead and ensure correct typing
        cdef np.ndarray[np.uint8_t, ndim=1] ticks = self.fvars['ticks'][idx]
        cdef np.ndarray[np.uint16_t, ndim=1] exp_data = self.fvars['exp_data'][idx]
        cdef np.ndarray[np.int64_t, ndim=3] nums = self.fvars['coeff_nums'][idx]
        cdef np.ndarray[np.uint64_t, ndim=3] denoms = self.fvars['coeff_denom'][idx]
        cdef np.ndarray[np.int8_t, ndim=1] modified = self.fvars['modified'][idx]
        cdef list digits
        # Initialize denominators to 1
        denoms[:, :, 0] = 1
        cum = 0
        i = 0
        for exp, coeff_tup in fvar:
            # Handle constant coefficient
            if exp._nonzero > 0:
                ticks[i] = exp._nonzero
            else:
                ticks[i] = -1
            for j in range(2*exp._nonzero):
                exp_data[cum] = exp._data[j]
                cum += 1
            k = 0
            for r in coeff_tup:
                num, denom = r.as_integer_ratio()
                if abs(num) > 2**63 or denom > 2**63:
                    print("Large integers encountered in FvarsHandler", num, denom)
                if abs(num) < 2**63:
                    nums[i, k, 0] = num
                else:
                    digits = num.digits(2**63)
                    # assert len(digits) <= self.bytes // 8, "Numerator {} is too large for shared FvarsHandler. Use at least {} bytes...".format(num, num.nbits()//8+1)
                    for t in range(len(digits)):
                        nums[i, k, t] = <np.int64_t>digits[t]
                if denom < 2**64:
                    denoms[i, k, 0] = denom
                else:
                    digits = denom.digits(2**64)
                    # assert len(digits) <= self.bytes // 8, "Denominator {} is too large for shared FvarsHandler. Use at least {} bytes...".format(denom, denom.nbits()//8+1)
                    for t in range(len(digits)):
                        denoms[i, k, t] = <np.uint64_t>digits[t]
                k += 1
            i += 1
        modified[:] = 1

    def __reduce__(self):
        r"""
        Provide pickling / unpickling support for ``self``.

        TESTS::

            sage: f = FusionRing("F4", 1).get_fmatrix()
            sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
            sage: n = f._poly_ring.ngens()
            sage: f.start_worker_pool()
            sage: n_proc = f.pool._processes
            sage: pids_name = f._pid_list.shm.name
            sage: fvars = FvarsHandler(n, f._field, f._idx_to_sextuple, init_data=f._fvars, use_mp=n_proc, pids_name=pids_name)
            sage: for s, fvar in loads(dumps(fvars)).items():
            ....:     assert f._fvars[s] == f._tup_to_fpoly(fvar)
            ....:
            sage: fvars.shm.unlink()
            sage: f.shutdown_worker_pool()
        """
        n = self.fvars.size
        idx_map = {i: s for s, i in self.sext_to_idx.items()}
        d = {s: fvar for s, fvar in self.items()}
        return make_FvarsHandler, (n, self.field, idx_map, d)

    def items(self):
        r"""
        Iterate through key-value pairs in the data structure as if it
        were a Python dict.

        As in a Python dict, the key-value pairs are yielded in no particular
        order.

        EXAMPLES::

            sage: f = FusionRing("G2", 1).get_fmatrix(inject_variables=True, new=True)
            creating variables fx1..fx5
            Defining fx0, fx1, fx2, fx3, fx4
            sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
            sage: shared_fvars = FvarsHandler(5, f._field, f._idx_to_sextuple, init_data=f._fvars)
            sage: for sextuple, fvar in shared_fvars.items():
            ....:     if sextuple == (f1, f1, f1, f1, f1, f1):
            ....:         f._tup_to_fpoly(fvar)
            ....:
            fx4
        """
        for sextuple in self.sext_to_idx:
            yield sextuple, self[sextuple]


def make_FvarsHandler(n, field, idx_map, init_data):
    r"""
    Provide pickling / unpickling support for :class:`FvarsHandler`.

    TESTS::

        sage: f = FusionRing("G2", 1).get_fmatrix()
        sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
        sage: n = f._poly_ring.ngens()
        sage: f.start_worker_pool()
        sage: n_proc = f.pool._processes
        sage: pids_name = f._pid_list.shm.name
        sage: fvars = FvarsHandler(n, f._field, f._idx_to_sextuple, init_data=f._fvars, use_mp=n_proc, pids_name=pids_name)
        sage: for s, fvar in loads(dumps(fvars)).items():            # indirect doctest
        ....:     assert f._fvars[s] == f._tup_to_fpoly(fvar)
        ....:
        sage: fvars.shm.unlink()
        sage: f.shutdown_worker_pool()
    """
    return FvarsHandler(n, field, idx_map, init_data=init_data)



# sage_setup: distribution = sagemath-singular
# ############################################################################
#
#       Copyright (C) 2011 Simon King <simon.king@uni-jena.de>
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or any later version.  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
#
# ############################################################################

from sage.structure.element cimport AlgebraElement, ModuleElement, Element
from sage.rings.polynomial.multi_polynomial_libsingular cimport MPolynomialRing_libsingular, MPolynomial_libsingular
from sage.algebras.letterplace.free_algebra_letterplace cimport FreeAlgebra_letterplace

cdef class FreeAlgebraElement_letterplace(AlgebraElement):
    cdef MPolynomial_libsingular _poly
    cpdef _add_(self, other)
    cpdef _mul_(self, other)



# sage_setup: distribution = sagemath-singular
"""
Highlevel functions for managing options of the libSingular interface

AUTHOR:

- Martin Albrecht
"""


class LibSingularGBDefaultContext:
    def __init__(self):
        """
        EXAMPLES::

            sage: from sage.libs.singular.standard_options import LibSingularGBDefaultContext
            sage: from sage.libs.singular.option import opt
            sage: P.<a,b,c> = PolynomialRing(QQ, 3, order='lex')
            sage: I = sage.rings.ideal.Katsura(P, 3)
            sage: I.gens()
            [a + 2*b + 2*c - 1, a^2 - a + 2*b^2 + 2*c^2, 2*a*b + 2*b*c - b]
            sage: opt['red_tail'] = False
            sage: opt['red_sb'] = False
            sage: import sage.libs.singular.function_factory
            sage: groebner = sage.libs.singular.function_factory.ff.groebner
            sage: rgb = groebner(I)
            sage: rgb
            [84*c^4 - 40*c^3 + c^2 + c, 7*b + 210*c^3 - 79*c^2 + 3*c, a + 2*b + 2*c - 1]
            sage: with LibSingularGBDefaultContext(): rgb = groebner(I)
            sage: rgb
            [84*c^4 - 40*c^3 + c^2 + c, 7*b + 210*c^3 - 79*c^2 + 3*c, 7*a - 420*c^3 + 158*c^2 + 8*c - 7]
        """
        from sage.libs.singular.option import opt_ctx
        self.libsingular_option_context = opt_ctx

    def __enter__(self):
        """
        EXAMPLES::

            sage: from sage.libs.singular.standard_options import LibSingularGBDefaultContext
            sage: from sage.libs.singular.option import opt
            sage: P.<a,b,c,d> = PolynomialRing(QQ, 4, order='lex')
            sage: I = sage.rings.ideal.Katsura(P, 4)
            sage: I.gens()
            [a + 2*b + 2*c + 2*d - 1, a^2 - a + 2*b^2 + 2*c^2 + 2*d^2, 2*a*b + 2*b*c - b + 2*c*d, 2*a*c + b^2 + 2*b*d - c]

            sage: opt['red_tail'] = False
            sage: opt['red_sb'] = False
            sage: import sage.libs.singular.function_factory
            sage: groebner = sage.libs.singular.function_factory.ff.groebner
            sage: rgb = groebner(I)
            sage: rgb
            [128304*d^8 - 93312*d^7 + 15552*d^6 + 3144*d^5 - 1120*d^4 + 36*d^3 + 15*d^2 - d,
             5913075*c + 371438283744*d^7 - 237550027104*d^6 + 22645939824*d^5 + 11520686172*d^4 - 2024910556*d^3 - 132524276*d^2 + 30947828*d,
             1044*b + 10976*c^3 + 67914*c^2*d - 9709*c^2 + 100296*c*d^2 - 32862*c*d + 2412*c + 64302*d^3 - 29483*d^2 + 2683*d,
             a + 2*b + 2*c + 2*d - 1]

            sage: with LibSingularGBDefaultContext(): rgb = groebner(I)
            sage: rgb
            [128304*d^8 - 93312*d^7 + 15552*d^6 + 3144*d^5 - 1120*d^4 + 36*d^3 + 15*d^2 - d,
             5913075*c + 371438283744*d^7 - 237550027104*d^6 + 22645939824*d^5 + 11520686172*d^4 - 2024910556*d^3 - 132524276*d^2 + 30947828*d,
             1971025*b - 97197721632*d^7 + 73975630752*d^6 - 12121915032*d^5 - 2760941496*d^4 + 814792828*d^3 - 1678512*d^2 - 9158924*d,
             5913075*a - 159690237696*d^7 + 31246269696*d^6 + 27439610544*d^5 - 6475723368*d^4 - 838935856*d^3 + 275119624*d^2 + 4884038*d - 5913075]
        """
        self.libsingular_option_context.__enter__()
        self.libsingular_option_context.opt.reset_default()
        self.libsingular_option_context.opt['red_sb'] = True
        self.libsingular_option_context.opt['red_tail'] = True
        self.libsingular_option_context.opt['deg_bound'] = 0
        self.libsingular_option_context.opt['mult_bound'] = 0

    def __exit__(self, typ, value, tb):
        """
        EXAMPLES::

            sage: from sage.libs.singular.standard_options import LibSingularGBDefaultContext
            sage: from sage.libs.singular.option import opt
            sage: P.<a,b,c,d> = PolynomialRing(GF(7), 4, order='lex')
            sage: I = sage.rings.ideal.Katsura(P, 4)
            sage: I.gens()
            [a + 2*b + 2*c + 2*d - 1, a^2 - a + 2*b^2 + 2*c^2 + 2*d^2, 2*a*b + 2*b*c - b + 2*c*d, 2*a*c + b^2 + 2*b*d - c]

            sage: opt['red_tail'] = False
            sage: opt['red_sb'] = False
            sage: import sage.libs.singular.function_factory
            sage: groebner = sage.libs.singular.function_factory.ff.groebner
            sage: rgb = groebner(I)
            sage: rgb
            [d^7 + 3*d^6 - d^5 + 3*d^4 + d^3 - d^2 + 3*d,
             c - 3*d^6 + 3*d^5 - d^4 + d^3 - 2*d,
             b + 3*c*d - 3*c + d^2 + 2*d,
             a + 2*b + 2*c + 2*d - 1]
            sage: with LibSingularGBDefaultContext(): rgb = groebner(I)
            sage: rgb
             [d^7 + 3*d^6 - d^5 + 3*d^4 + d^3 - d^2 + 3*d,
              c - 3*d^6 + 3*d^5 - d^4 + d^3 - 2*d,
              b - 3*d^6 + 2*d^4 + d^3 + 2*d^2 - 3*d,
              a - 2*d^6 + d^5 - 2*d^4 + 3*d^3 + 3*d^2 - 2*d - 1]
        """
        self.libsingular_option_context.__exit__(typ, value, tb)


def libsingular_gb_standard_options(func):
    r"""
    Decorator to force a reduced Singular groebner basis.

    TESTS::

        sage: P.<a,b,c,d,e> = PolynomialRing(GF(127))
        sage: J = sage.rings.ideal.Cyclic(P).homogenize()
        sage: from sage.misc.sageinspect import sage_getsource
        sage: "basis.reduced" in sage_getsource(J.interreduced_basis)
        True

    The following tests against a bug that was fixed in :issue:`11298`::

        sage: from sage.misc.sageinspect import sage_getsourcelines, sage_getargspec
        sage: P.<x,y> = QQ[]
        sage: I = P*[x,y]
        sage: sage_getargspec(I.interreduced_basis)
        FullArgSpec(args=['self'], varargs=None, varkw=None, defaults=None,
                    kwonlyargs=[], kwonlydefaults=None, annotations={})
        sage: sage_getsourcelines(I.interreduced_basis)
        (['    @handle_AA_and_QQbar\n',
          '    @singular_gb_standard_options\n',
          '    @libsingular_gb_standard_options\n',
          '    def interreduced_basis(self):\n',
          ...
          '        return self.basis.reduced()\n'], ...)

    .. NOTE::

       This decorator is used automatically internally so the user
       does not need to use it manually.
    """
    from sage.misc.decorators import sage_wraps

    @sage_wraps(func)
    def wrapper(*args, **kwds):
        """
        Execute function in ``LibSingularGBDefaultContext``.
        """
        with LibSingularGBDefaultContext():
            return func(*args, **kwds)
    return wrapper



# sage_setup: distribution = sagemath-singular
"""
libSingular: Function Factory

AUTHORS:

- Martin Albrecht (2010-01): initial version
"""
# ****************************************************************************
#       Copyright (C) 2010 Martin Albrecht <M.R.Albrecht@rhul.ac.uk>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.libs.singular.function import singular_function, lib, list_of_functions


class SingularFunctionFactory:
    """
    A convenient interface to libsingular functions.
    """
    def __getattr__(self, name):
        """
        EXAMPLES::

            sage: import sage.libs.singular.function_factory
            sage: groebner = sage.libs.singular.function_factory.ff.groebner
            sage: groebner
            groebner (singular function)

            sage: import sage.libs.singular.function_factory
            sage: primdecSY = sage.libs.singular.function_factory.ff.primdec__lib.primdecSY
            sage: primdecSY
            primdecSY (singular function)
        """
        if name.startswith("_"):
            raise AttributeError("Singular Function Factory has no attribute '%s'" % name)

        try:
            return singular_function(name)
        except NameError:
            if name.endswith("__lib"):
                name = name[:-5]
                lib(name + ".lib")
                return SingularFunctionFactory()
            else:
                raise NameError("function or package '%s' unknown." % (name))

    def __dir__(self):
        """
        EXAMPLES::

            sage: import sage.libs.singular.function_factory
            sage: "groebner" in sage.libs.singular.function_factory.ff.__dir__()
            True
        """
        return list_of_functions()


ff = SingularFunctionFactory()



# sage_setup: distribution = sagemath-singular
# sage.doctest: needs sage.combinat sage.groups
"""
Fast F-Matrix Methods
"""
# ****************************************************************************
#  Copyright (C) 2021 Guillermo Aboumrad <gh_willieab>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

cimport cython
from sage.algebras.fusion_rings.poly_tup_engine cimport (
    compute_known_powers,
    get_variables_degrees, variables,
    poly_to_tup, _tup_to_poly,
    subs, subs_squares, reduce_poly_dict, resize,
    _flatten_coeffs, _unflatten_coeffs,
    has_appropriate_linear_term,
    resize
)
from sage.algebras.fusion_rings.shm_managers cimport KSHandler, FvarsHandler
from sage.rings.number_field.number_field_element cimport NumberFieldElement_absolute
from sage.rings.polynomial.multi_polynomial_libsingular cimport MPolynomial_libsingular, MPolynomialRing_libsingular
from sage.rings.polynomial.polydict cimport ETuple

from ctypes import cast, py_object
from itertools import product
from sage.rings.ideal import Ideal
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

##########################
#   Fast class methods   #
##########################

cpdef _solve_for_linear_terms(factory, list eqns=None):
    r"""
    Solve for a linear term occurring in a two-term equation, and for
    variables appearing in univariate single-term equations.

    EXAMPLES::

        sage: f = FusionRing("D3", 1).get_fmatrix(inject_variables=True, new=True)
        creating variables fx1..fx27
        Defining fx0, ..., fx26
        sage: f._reset_solver_state()
        sage: f.ideal_basis = [fx0**3, fx0 + fx3**4, fx2**2 - fx3, fx2 - fx3**2, fx4 - fx2]
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: f.ideal_basis = [poly_to_tup(p) for p in f.ideal_basis]
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_tup_sortkey
        sage: f.ideal_basis.sort(key=poly_tup_sortkey)
        sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
        sage: n = f._poly_ring.ngens()
        sage: f._fvars = FvarsHandler(n, f._field, f._idx_to_sextuple, init_data=f._fvars)
        sage: from sage.algebras.fusion_rings.fast_parallel_fmats_methods import _solve_for_linear_terms
        sage: _solve_for_linear_terms(f)
        True
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[0]])
        0
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[2]])
        fx4
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[3]])
        fx3
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[4]])
        fx4
    """
    if eqns is None:
        eqns = factory.ideal_basis

    cdef bint linear_terms_exist = False
    cdef ETuple exp, rhs_exp
    cdef int max_var
    cdef tuple eq_tup
    cdef FvarsHandler fvars = factory._fvars
    cdef NumberFieldElement_absolute coeff, other
    cdef tuple rhs_coeff
    for eq_tup in eqns:
        if len(eq_tup) == 1:
            vars = variables(eq_tup)
            if len(vars) == 1 and not factory._solved[vars[0]]:
                fvars[factory._idx_to_sextuple[vars[0]]] = tuple()
                factory._solved[vars[0]] = True
                linear_terms_exist = True

                # TESTS:
                # s = factory._idx_to_sextuple[vars[0]]
                # factory.test_fvars[s] = tuple()
                # assert factory.test_fvars[s] == fvars[s], "OG value {}, Shared: {}".format(fvars[s], factory.test_fvars[s])
        if len(eq_tup) == 2:
            idx = has_appropriate_linear_term(eq_tup)
            if idx < 0:
                continue
            # The chosen term is guaranteed to be univariate in the largest variable
            exp = eq_tup[idx][0]
            max_var = exp._data[0]
            if not factory._solved[max_var]:
                rhs_exp = eq_tup[(idx+1) % 2][0]
                coeff = factory._field(list(eq_tup[(idx+1) % 2][1]))
                other = factory._field(list(eq_tup[idx][1]))
                rhs_coeff = tuple((-coeff / other)._coefficients())
                fvars[factory._idx_to_sextuple[max_var]] = ((rhs_exp, rhs_coeff), )
                factory._solved[max_var] = True
                linear_terms_exist = True

                # TESTS:
                # s = factory._idx_to_sextuple[max_var]
                # factory.test_fvars[s] = ((rhs_exp, rhs_coeff), )
                # assert _unflatten_coeffs(factory._field, factory.test_fvars[s]) == fvars[s], "OG value {}, Shared: {}".format(factory.test_fvars[s], fvars[s])
    return linear_terms_exist

cpdef _backward_subs(factory, bint flatten=True):
    r"""
    Perform backward substitution on ``self.ideal_basis``, traversing
    variables in reverse lexicographical order.

    EXAMPLES::

        sage: f = FusionRing("D3", 1).get_fmatrix(inject_variables=True, new=True)
        creating variables fx1..fx27
        Defining fx0, ..., fx26
        sage: f._reset_solver_state()
        sage: f.ideal_basis = [fx0**3, fx0 + fx3**4, fx2**2 - fx3, fx2 - fx3**2, fx4 - fx2]
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_to_tup
        sage: f.ideal_basis = [poly_to_tup(p) for p in f.ideal_basis]
        sage: from sage.algebras.fusion_rings.poly_tup_engine import poly_tup_sortkey
        sage: f.ideal_basis.sort(key=poly_tup_sortkey)
        sage: from sage.algebras.fusion_rings.shm_managers import FvarsHandler
        sage: n = f._poly_ring.ngens()
        sage: f._fvars = FvarsHandler(n, f._field, f._idx_to_sextuple, init_data=f._fvars)
        sage: from sage.algebras.fusion_rings.fast_parallel_fmats_methods import _solve_for_linear_terms
        sage: _solve_for_linear_terms(f)
        True
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[0]])
        0
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[2]])
        fx4
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[3]])
        fx3
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[4]])
        fx4
        sage: f.ideal_basis.append(poly_to_tup(fx4-1))
        sage: f.ideal_basis.sort(key=poly_tup_sortkey)
        sage: _solve_for_linear_terms(f)
        True
        sage: from sage.algebras.fusion_rings.fast_parallel_fmats_methods import _backward_subs
        sage: _backward_subs(f)
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[0]])
        0
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[2]])
        1
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[3]])
        fx3
        sage: f._tup_to_fpoly(f._fvars[f._idx_to_sextuple[4]])
        1
    """
    one = factory._field.one()
    fvars = factory._fvars
    solved = factory._solved
    cdef KSHandler _ks = factory._ks
    cdef dict idx_to_sextuple = factory._idx_to_sextuple
    cdef int nvars = len(idx_to_sextuple)
    for i in range(nvars-1, -1, -1):
        sextuple = idx_to_sextuple[i]
        rhs = fvars[sextuple]
        d = {var_idx: fvars[idx_to_sextuple[var_idx]]
             for var_idx in variables(rhs) if solved[var_idx]}
        if d:
            kp = compute_known_powers(get_variables_degrees([rhs], nvars), d, one)
            res = tuple(subs_squares(subs(rhs, kp, one), _ks).items())
            if flatten:
                res = _flatten_coeffs(res)
            fvars[sextuple] = res


cdef _fmat(fvars, _Nk_ij, id_anyon, a, b, c, d, x, y):
    """
    Cython version of fmat class method. Using cdef for fastest dispatch
    """
    if _Nk_ij(a, b, x) == 0 or _Nk_ij(x, c, d) == 0 or _Nk_ij(b, c, y) == 0 or _Nk_ij(a, y, d) == 0:
        return 0
    # Some known F-symbols
    if a == id_anyon:
        return int(x == b and y == d)
    if b == id_anyon:
        return int(x == a and y == c)
    if c == id_anyon:
        return int(x == d and y == b)
    return fvars[a, b, c, d, x, y]

######################################
#   Fast fusion coefficients cache   #
######################################

# from sage.misc.cachefunc import cached_function
# cdef dict _Nk_ij = dict()

# cpdef _Nk_ij(factory, proc):
# cdef int coeff
# for a, b, c in product(factory._FR.basis(), repeat=3):
# try:
# coeff = (a*b).monomial_coefficients(copy=False)[c.weight()]
# except:
# coeff = 0
# _Nk_ij[a, b, c] = coeff

# cpdef int _Nk_ij(a, b, c):
# try:
# return (a*b).monomial_coefficients(copy=False)[c.weight()]
# except KeyError:
# return 0
#
# _Nk_ij = cached_function(_Nk_ij, name='_Nk_ij')

###############
#   Mappers   #
###############

cdef req_cy(tuple basis, r_matrix, dict fvars, Nk_ij, id_anyon, tuple sextuple):
    """
    Given an FMatrix factory and a sextuple, return a hexagon equation
    as a polynomial object.
    """
    a, b, c, d, e, g = sextuple
    # To add typing we need to ensure all fmats.fmat are of the same type?
    # Return fmats._poly_ring.zero() and fmats._poly_ring.one() instead of 0 and 1?
    lhs = r_matrix(a, c, e, base_coercion=False) * _fmat(fvars, Nk_ij, id_anyon, a, c, b, d, e, g) * r_matrix(b, c, g, base_coercion=False)
    rhs = 0
    for f in basis:
        rhs += _fmat(fvars, Nk_ij, id_anyon, c, a, b, d, e, f) * r_matrix(f, c, d, base_coercion=False) * _fmat(fvars, Nk_ij, id_anyon, a, b, c, d, f, g)
    return lhs-rhs


@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef get_reduced_hexagons(factory, tuple mp_params):
    """
    Set up and reduce the hexagon equations corresponding to this worker.
    """
    # Set up multiprocessing parameters
    cdef list worker_results = list()
    cdef int child_id, n_proc
    cdef unsigned long i
    child_id, n_proc, output = mp_params
    cdef tuple sextuple, red

    # Pre-compute common parameters for speed
    cdef tuple basis = tuple(factory._FR.basis())
    cdef dict fvars
    cdef bint must_zip_up = False
    if not output:
        fvars = {s: factory._poly_ring.gen(i) for i, s in factory._idx_to_sextuple.items()}
    else:
        # Handle both cyclotomic and orthogonal solution method
        for k, v in factory._fvars.items():
            must_zip_up = isinstance(v, tuple)
            break
        if must_zip_up:
            fvars = {k: _tup_to_poly(v, parent=factory._poly_ring) for k, v in factory._fvars.items()}
        else:
            fvars = factory._fvars
    r_matrix = factory._FR.r_matrix
    _Nk_ij = factory._FR.Nk_ij
    id_anyon = factory._FR.one()
    _field = factory._field
    cdef NumberFieldElement_absolute one = _field.one()
    cdef ETuple _nnz = factory._nnz
    _ks = factory._ks

    # Computation loop
    it = product(basis, repeat=6)
    for i in range(len(basis)**6):
        sextuple = next(it)
        if i % n_proc == child_id:
            he = req_cy(basis, r_matrix, fvars, _Nk_ij, id_anyon, sextuple)
            if he:
                red = reduce_poly_dict(he.monomial_coefficients(), _nnz, _ks, one)

                # Avoid pickling cyclotomic coefficients
                red = _flatten_coeffs(red)

                worker_results.append(red)

    return collect_eqns(worker_results)

cdef MPolynomial_libsingular feq_cy(tuple basis, fvars, Nk_ij, id_anyon, zero, tuple nonuple, bint prune=False):
    r"""
    Given an FMatrix factory and a nonuple, return a pentagon equation
    as a polynomial object.
    """
    a, b, c, d, e, f, g, k, l = nonuple
    lhs = _fmat(fvars, Nk_ij, id_anyon, f, c, d, e, g, l) * _fmat(fvars, Nk_ij, id_anyon, a, b, l, e, f, k)
    if lhs == 0 and prune:
        # it is believed that if lhs=0, the equation carries no new information
        return zero
    rhs = zero
    for h in basis:
        rhs += _fmat(fvars, Nk_ij, id_anyon, a, b, c, g, f, h)*_fmat(fvars, Nk_ij, id_anyon, a, h, d, e, g, k)*_fmat(fvars, Nk_ij, id_anyon, b, c, d, k, h, l)
    return lhs - rhs


@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef get_reduced_pentagons(factory, tuple mp_params):
    r"""
    Set up and reduce the pentagon equations corresponding to this worker.
    """
    # Set up multiprocessing parameters
    cdef list worker_results = list()
    cdef int child_id, n_proc
    child_id, n_proc, _ = mp_params
    cdef unsigned long i
    cdef tuple nonuple, red
    cdef MPolynomial_libsingular pe

    # Pre-compute common parameters for speed
    cdef tuple basis = tuple(factory._FR.basis())
    # Handle both cyclotomic and orthogonal solution method
    cdef bint must_zip_up
    for k, v in factory._fvars.items():
        must_zip_up = isinstance(v, tuple)
        break
    cdef dict fvars
    if must_zip_up:
        fvars = {k: _tup_to_poly(v, parent=factory._poly_ring) for k, v in factory._fvars.items()}
    else:
        fvars = factory._fvars
    _Nk_ij = factory._FR.Nk_ij
    id_anyon = factory._FR.one()
    _field = factory._field
    cdef NumberFieldElement_absolute one = _field.one()
    cdef MPolynomial_libsingular zero = factory._poly_ring.zero()
    cdef KSHandler _ks = factory._ks
    factory._nnz = factory._get_known_nonz()
    cdef ETuple _nnz = factory._nnz

    # Computation loop
    it = product(basis, repeat=9)
    for i in range(len(basis)**9):
        nonuple = next(it)
        if i % n_proc == child_id:
            pe = feq_cy(basis, fvars, _Nk_ij, id_anyon, zero, nonuple, prune=True)
            if pe:
                red = reduce_poly_dict(pe.monomial_coefficients(), _nnz, _ks, one)

                # Avoid pickling cyclotomic coefficients
                red = _flatten_coeffs(red)

                worker_results.append(red)
    return collect_eqns(worker_results)

cdef list update_reduce(factory, list eqns):
    r"""
    Substitute known values, known squares, and reduce.
    """
    cdef list res = list()
    cdef tuple eq_tup, red, unflat
    cdef dict eq_dict

    # Pre-compute common parameters for speed
    _field = factory._field
    one = _field.one()
    cdef KSHandler _ks = factory._ks
    # Update reduction params
    factory._nnz = factory._get_known_nonz()
    factory._kp = compute_known_powers(factory._var_degs, factory._get_known_vals(), factory._field.one())
    cdef dict _kp = factory._kp
    cdef ETuple _nnz = factory._nnz

    for i in range(len(eqns)):
        eq_tup = eqns[i]
        # Construct cyclotomic field elts from list repn
        unflat = _unflatten_coeffs(_field, eq_tup)

        eq_dict = subs(unflat, _kp, one)
        red = reduce_poly_dict(eq_dict, _nnz, _ks, one)

        # Avoid pickling cyclotomic coefficients
        red = _flatten_coeffs(red)

        res.append(red)
    return collect_eqns(res)

cdef list compute_gb(factory, tuple args):
    r"""
    Compute the reduced Groebner basis for given equations iterable.
    """
    cdef list res = list()
    cdef list eqns, sorted_vars
    eqns, term_order = args
    # Define smaller poly ring in component vars
    sorted_vars = []
    cdef tuple eq_tup
    cdef int fx
    for eq_tup in eqns:
        for fx in variables(eq_tup):
            sorted_vars.append(fx)
    sorted_vars = sorted(set(sorted_vars))
    cdef MPolynomialRing_libsingular R = PolynomialRing(factory._FR.field(), len(sorted_vars), 'a', order=term_order)

    # Zip tuples into R and compute Groebner basis
    cdef idx_map = {old: new for new, old in enumerate(sorted_vars)}
    nvars = len(sorted_vars)
    F = factory.field()
    cdef list polys = list()
    for eq_tup in eqns:
        eq_tup = _unflatten_coeffs(F, eq_tup)
        polys.append(_tup_to_poly(resize(eq_tup, idx_map, nvars), parent=R))
    gb = Ideal(sorted(polys)).groebner_basis(algorithm='libsingular:slimgb')

    # Change back to fmats poly ring and append to temp_eqns
    cdef dict inv_idx_map = {v: k for k, v in idx_map.items()}
    cdef tuple t
    nvars = factory._poly_ring.ngens()
    for p in gb:
        t = resize(poly_to_tup(p), inv_idx_map, nvars)

        # Avoid pickling cyclotomic coefficients
        t = _flatten_coeffs(t)

        res.append(t)
    return collect_eqns(res)

################
#   Reducers   #
################

cdef inline list collect_eqns(list eqns):
    r"""
    Helper function for returning processed results back to parent process.

    Trivial reducer: simply collects objects with the same key in the worker.
    This method is only useful when called after :meth:`executor`, whose
    function argument appends output to the ``worker_results`` list.
    """
    # Discard the zero polynomial
    reduced = set(eqns) - set([tuple()])
    return list(reduced)

##############################
#   Parallel code executor   #
##############################

# Hard-coded module __dict__-style attribute with visible cdef methods
cdef dict mappers = {
    "get_reduced_hexagons": get_reduced_hexagons,
    "get_reduced_pentagons": get_reduced_pentagons,
    "update_reduce": update_reduce,
    "compute_gb": compute_gb,
    "pent_verify": pent_verify
    }

cpdef executor(tuple params):
    r"""
    Execute a function defined in this module
    (``sage.algebras.fusion_rings.fast_parallel_fmats_methods``) in a worker
    process, and supply the factory parameter by constructing a reference
    to the ``FMatrix`` object in the worker's memory address space from
    its ``id``.

    INPUT:

    - ``params`` -- tuple ``((fn_name, fmats_id), fn_args)`` where
      ``fn_name`` is the name of the function to be executed, ``fmats_id``
      is the ``id`` of the :class:`FMatrix` object, and ``fn_args`` is a
      tuple containing all arguments to be passed to the function ``fn_name``.

    .. NOTE::

        When the parent process is forked, each worker gets a copy of
        every  global variable. The virtual memory address of object `X` in
        the parent process equals the *virtual* memory address of the copy
        of object `X` in each worker, so we may construct references to
        forked copies of `X` using an ``id`` obtained in the parent process.

    TESTS::

        sage: from sage.algebras.fusion_rings.fast_parallel_fmats_methods import executor
        sage: fmats = FusionRing("A1", 3).get_fmatrix()
        sage: fmats._reset_solver_state()
        sage: params = (('get_reduced_hexagons', id(fmats)), (0, 1, True))
        sage: len(executor(params)) == 63
        True
        sage: fmats = FusionRing("E6", 1).get_fmatrix()
        sage: fmats._reset_solver_state()
        sage: params = (('get_reduced_hexagons', id(fmats)), (0, 1, False))
        sage: len(executor(params)) == 6
        True
    """
    (fn_name, fmats_id), args = params
    # Construct a reference to global FMatrix object in this worker's memory
    fmats_obj = cast(fmats_id, py_object).value
    # Bind module method to FMatrix object in worker process, and call the method
    return mappers[fn_name](fmats_obj, args)

####################
#   Verification   #
####################

cdef feq_verif(factory, worker_results, fvars, Nk_ij, id_anyon, tuple nonuple, float tol=5e-8):
    r"""
    Check the pentagon equation corresponding to the given nonuple.
    """
    a, b, c, d, e, f, g, k, l = nonuple
    cdef float diff, lhs, rhs

    lhs = _fmat(fvars, Nk_ij, id_anyon, f, c, d, e, g, l)*_fmat(fvars, Nk_ij, id_anyon, a, b, l, e, f, k)
    rhs = 0.0
    for h in factory._FR.basis():
        rhs += _fmat(fvars, Nk_ij, id_anyon, a, b, c, g, f, h)*_fmat(fvars, Nk_ij, id_anyon, a, h, d, e, g, k)*_fmat(fvars, Nk_ij, id_anyon, b, c, d, k, h, l)
    diff = lhs - rhs
    if diff > tol or diff < -tol:
        worker_results.append(diff)


@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef pent_verify(factory, tuple mp_params):
    r"""
    Generate all the pentagon equations assigned to this process,
    and reduce them.
    """
    child_id, n_proc, verbose = mp_params
    cdef tuple nonuple
    cdef unsigned long long i
    cdef list worker_results = list()

    # Pre-compute common parameters for speed
    Nk_ij = factory._FR.Nk_ij
    cdef dict fvars = factory._fvars
    id_anyon = factory._FR.one()
    for i, nonuple in enumerate(product(factory._FR.basis(), repeat=9)):
        if i % n_proc == child_id:
            feq_verif(factory, worker_results, fvars, Nk_ij, id_anyon, nonuple)
        if i % 50000000 == 0 and i and verbose:
            print("{:5d}m equations checked... {} potential misses so far...".format(i // 1000000, len(worker_results)))
